@@ -18,11 +18,11 @@ Params::Validate::Strict - Validates a set of parameters against a schema
 
 =head1 VERSION
 
-Version 0.19
+Version 0.20
 
 =cut
 
-our $VERSION = '0.19';
+our $VERSION = '0.20';
 
 =head1 SYNOPSIS
 
@@ -46,14 +46,39 @@ our $VERSION = '0.19';
         print "Example 1: Validation failed: $@\n";
     }
 
+Upon first reading this may seem overly complex and full of scope creep in a sledgehammer to crack a nut sort of way,
+however two use cases make use of the extensive logic that comes with this code
+and I have a couple of other reasons for writing it.
+
+=over 4
+
+=item * Black Box Testing
+
+The schema can be plumbed into L<App::Test::Generator> to automatically create a set of black-box test cases.
+
+=item * WAF
+
+The schema can be plumbed into a WAF to protect from random user input.
+
+=item * Improved API Documentation
+
+Even if you don't use this module,
+the specification syntax can help with documentation.
+
+=item * I like it
+
+I find it fun to write this,
+even if nobody else finds it useful,
+though I hope you will.
+
+=back
+
 =head1	METHODS
 
 =head2 validate_strict
 
 Validates a set of parameters against a schema.
 
-The schema can then be plumbed into L<App::Test::Generator> to automatically create a set of black-box test cases.
-
 This function takes two mandatory arguments:
 
 =over 4
@@ -70,7 +95,7 @@ The keys of the hash are the parameter names, and the values are the parameter v
 
 =back
 
-It takes two optional arguments:
+It takes three optional arguments:
 
 =over 4
 
@@ -83,6 +108,60 @@ It must be one of C<die> (the default), C<warn>, or C<ignore>.
 
 A logging object that understands messages such as C<error> and C<warn>.
 
+=item * C<custom_types>
+
+A reference to a hash that defines reusable custom types.
+Custom types allow you to define validation rules once and reuse them throughout your schema,
+making your validation logic more maintainable and readable.
+
+Each custom type is defined as a hash reference containing the same validation rules available for regular parameters
+(C<type>, C<min>, C<max>, C<matches>, C<memberof>, C<callback>, etc.).
+
+  my $custom_types = {
+    email => {
+      type => 'string',
+      matches => qr/^[\w\.\-]+@[\w\.\-]+\.\w+$/,
+      error_message => 'Invalid email address format'
+    }, phone => {
+      type => 'string',
+      matches => qr/^\+?[1-9]\d{1,14}$/,
+      min => 10,
+      max => 15
+    }, percentage => {
+      type => 'number',
+      min => 0,
+      max => 100
+    }, status => {
+      type => 'string',
+      memberof => ['draft', 'published', 'archived']
+    }
+  };
+
+  my $schema = {
+    user_email => { type => 'email' },
+    contact_number => { type => 'phone', optional => 1 },
+    completion => { type => 'percentage' },
+    post_status => { type => 'status' }
+  };
+
+  my $validated = validate_strict(
+    schema => $schema,
+    input => $input,
+    custom_types => $custom_types
+  );
+
+Custom types can be extended or overridden in the schema by specifying additional constraints:
+
+  my $schema = {
+    admin_username => {
+      type => 'username',  # Uses custom type definition
+      min => 5,            # Overrides custom type's min value
+      max => 15            # Overrides custom type's max value
+    }
+  };
+
+Custom types work seamlessly with nested schemas, optional parameters, and all other validation features.
+
 =back
 
 The schema can define the following rules for each parameter:
@@ -183,14 +262,14 @@ The custom error message to be used in the event of a validation failure.
 You can validate nested hashrefs and arrayrefs using the C<schema> property:
 
     my $schema = {
-        user => {
+        user => {	# 'user' is a hashref
             type => 'hashref',
-            schema => {
+            schema => {	# Specify what the elements of the hash should be
                 name => { type => 'string' },
                 age => { type => 'integer', min => 0 },
-                hobbies => {
+                hobbies => {	# 'hobbies' is an array ref that this user has
                     type => 'arrayref',
-                    schema => { type => 'string' }, # Validate each element
+                    schema => { type => 'string' }, # Validate each hobby
                     min => 1 # At least one hobby
                 }
             }
@@ -231,8 +310,123 @@ or undef if it's allowed.
       }
     };
 
+=item * C<transform>
+
+A code reference to a subroutine that transforms/sanitizes the parameter value before validation.
+The subroutine should accept the parameter value as an argument and return the transformed value.
+The transformation is applied before any validation rules are checked, allowing you to normalize
+or clean data before it is validated.
+
+Common use cases include trimming whitespace, normalizing case, formatting phone numbers,
+sanitizing user input, and converting between data formats.
+
+  # Simple string transformations
+  username => {
+    type => 'string',
+    transform => sub { lc(trim($_[0])) },  # lowercase and trim
+    matches => qr/^[a-z0-9_]+$/
+  }
+
+  email => {
+    type => 'string',
+    transform => sub { lc(trim($_[0])) },  # normalize email
+    matches => qr/^[\w\.\-]+@[\w\.\-]+\.\w+$/
+  }
+
+  # Array transformations
+  tags => {
+    type => 'arrayref',
+    transform => sub { [map { lc($_) } @{$_[0]}] },  # lowercase all elements
+    element_type => 'string'
+  }
+
+  keywords => {
+    type => 'arrayref',
+    transform => sub {
+      my @arr = map { lc(trim($_)) } @{$_[0]};
+      my %seen;
+      return [grep { !$seen{$_}++ } @arr];  # remove duplicates
+    }
+  }
+
+  # Numeric transformations
+  quantity => {
+    type => 'integer',
+    transform => sub { int($_[0] + 0.5) },  # round to nearest integer
+    min => 1
+  }
+
+  # Sanitization
+  slug => {
+    type => 'string',
+    transform => sub {
+      my $str = lc(trim($_[0]));
+      $str =~ s/[^\w\s-]//g;  # remove special characters
+      $str =~ s/\s+/-/g;      # replace spaces with hyphens
+      return $str;
+    },
+    matches => qr/^[a-z0-9-]+$/
+  }
+
+  phone => {
+    type => 'string',
+    transform => sub {
+      my $str = $_[0];
+      $str =~ s/\D//g;  # remove all non-digits
+      return $str;
+    },
+    matches => qr/^\d{10}$/
+  }
+
+The C<transform> function is applied to the value before any validation checks (C<min>, C<max>,
+C<matches>, C<callback>, etc.), ensuring that validation rules are checked against the cleaned data.
+
+Transformations work with all parameter types including nested structures:
+
+  user => {
+    type => 'hashref',
+    schema => {
+      name => {
+        type => 'string',
+        transform => sub { trim($_[0]) }
+      },
+      email => {
+        type => 'string',
+        transform => sub { lc(trim($_[0])) }
+      }
+    }
+  }
+
+Transformations can also be defined in custom types for reusability:
+
+  my $custom_types = {
+    email => {
+      type => 'string',
+      transform => sub { lc(trim($_[0])) },
+      matches => qr/^[\w\.\-]+@[\w\.\-]+\.\w+$/
+    }
+  };
+
+Note that the transformed value is what gets returned in the validated result and is what
+subsequent validation rules will check against. If a transformation might fail, ensure it
+handles edge cases appropriately.
+It is the responsibilty of the transformer to ensure that the type of the returned value is correct,
+since that is what will be validated.
+
 =back
 
+Many validators also allow a code ref to be passed so that you can create your own, conditional validation rule, e.g.:
+
+  $schema = {
+    age => {
+      type => 'integer',
+      min => sub {
+          my ($value, $all_params) = @_;
+          return $all_params->{country} eq 'US' ? 21 : 18;
+      }
+    }
+  }
+
 If a parameter is optional and its value is C<undef>,
 validation will be skipped for that parameter.
 
@@ -281,6 +475,7 @@ sub validate_strict
 	my $args = $params->{'args'} || $params->{'input'};
 	my $unknown_parameter_handler = $params->{'unknown_parameter_handler'} || 'die';
 	my $logger = $params->{'logger'};
+	my $custom_types = $params->{'custom_types'};
 
 	# Check if schema and args are references to hashes
 	if(ref($schema) ne 'HASH') {
@@ -326,8 +521,27 @@ sub validate_strict
 			$rules = { type => $rules };
 		}
 
+		my $is_optional = 0;
+
+		if(ref($rules) eq 'HASH') {
+			if($rules->{'transform'} && defined($value)) {
+				if(ref($rules->{'transform'}) eq 'CODE') {
+					$value = &{$rules->{'transform'}}($value);
+				} else {
+					_error($logger, 'validate_strict: transforms must be a code ref');
+				}
+			}
+			if(exists($rules->{optional})) {
+				if(ref($rules->{'optional'}) eq 'CODE') {
+					$is_optional = &{$rules->{optional}}($value, $args);
+				} else {
+					$is_optional = $rules->{'optional'};
+				}
+			}
+		}
+
 		# Handle optional parameters
-		if((ref($rules) eq 'HASH') && $rules->{optional}) {
+		if((ref($rules) eq 'HASH') && $is_optional) {
 			if(!exists($args->{$key})) {
 				if(exists($rules->{'default'})) {
 					# Populate missing optional parameters with the specified output values
@@ -360,6 +574,10 @@ sub validate_strict
 			foreach my $rule_name (keys %$rules) {
 				my $rule_value = $rules->{$rule_name};
 
+				if((ref($rule_value) eq 'CODE') && ($rule_name ne 'validate') && ($rule_name ne 'callback')) {
+					$rule_value = &{$rule_value}($value, $args);
+				}
+
 				if($rule_name eq 'type') {
 					my $type = lc($rule_value);
 
@@ -461,6 +679,17 @@ sub validate_strict
 								_error($logger, "validate_strict: Parameter '$key' must be an object");
 							}
 						}
+					} elsif(my $custom_type = $custom_types->{$type}) {
+						if($custom_type->{'transform'}) {
+							# The custom type has a transform embedded within it
+							if(ref($custom_type->{'transform'}) eq 'CODE') {
+								$value = &{$custom_type->{'transform'}}($value);
+							} else {
+								_error($logger, 'validate_strict: transforms must be a code ref');
+								next;
+							}
+						}
+						validate_strict({ input => { $key => $value }, schema => { $key => $custom_type }, custom_types => $custom_types });
 					} else {
 						_error($logger, "validate_strict: Unknown type '$type'");
 					}
@@ -468,7 +697,12 @@ sub validate_strict
 					if(!defined($rules->{'type'})) {
 						_error($logger, "validate_strict: Don't know type of '$key' to determine its minimum value $rule_value");
 					}
-					if($rules->{'type'} eq 'string') {
+					my $type = lc($rules->{'type'});
+					if(exists($custom_types->{$type}->{'min'})) {
+						$rule_value = $custom_types->{$type}->{'min'};
+						$type = $custom_types->{$type}->{'type'};
+					}
+					if($type eq 'string') {
 						if(!defined($value)) {
 							next;	# Skip if string is undefined
 						}
@@ -508,7 +742,7 @@ sub validate_strict
 								_error($logger, "validate_strict: Parameter '$key' must contain at least $rule_value keys");
 							}
 						}
-					} elsif(($rules->{'type'} eq 'integer') || ($rules->{'type'} eq 'number') || ($rules->{'type'} eq 'float')) {
+					} elsif(($type eq 'integer') || ($type eq 'number') || ($type eq 'float')) {
 						if(!defined($value)) {
 							next;	# Skip if hash is undefined
 						}
@@ -520,13 +754,18 @@ sub validate_strict
 							}
 						}
 					} else {
-						_error($logger, "validate_strict: Parameter '$key' has meaningless min value $rule_value");
+						_error($logger, "validate_strict: Parameter '$key' of type '$type' has meaningless min value $rule_value");
 					}
 				} elsif($rule_name eq 'max') {
 					if(!defined($rules->{'type'})) {
 						_error($logger, "validate_strict: Don't know type of '$key' to determine its maximum value $rule_value");
 					}
-					if($rules->{'type'} eq 'string') {
+					my $type = lc($rules->{'type'});
+					if(exists($custom_types->{$type}->{'max'})) {
+						$rule_value = $custom_types->{$type}->{'max'};
+						$type = $custom_types->{$type}->{'type'};
+					}
+					if($type eq 'string') {
 						if(!defined($value)) {
 							next;	# Skip if string is undefined
 						}
@@ -566,7 +805,7 @@ sub validate_strict
 								_error($logger, "validate_strict: Parameter '$key' must contain no more than $rule_value keys");
 							}
 						}
-					} elsif(($rules->{'type'} eq 'integer') || ($rules->{'type'} eq 'number') || ($rules->{'type'} eq 'float')) {
+					} elsif(($type eq 'integer') || ($type eq 'number') || ($type eq 'float')) {
 						if(!defined($value)) {
 							next;	# Skip if hash is undefined
 						}
@@ -586,7 +825,7 @@ sub validate_strict
 							}
 						}
 					} else {
-						_error($logger, "validate_strict: Parameter '$key' has meaningless max value $rule_value");
+						_error($logger, "validate_strict: Parameter '$key' of type '$type' has meaningless max value $rule_value");
 					}
 				} elsif($rule_name eq 'matches') {
 					if(!defined($value)) {
@@ -640,21 +879,22 @@ sub validate_strict
 						next;	# Skip if string is undefined
 					}
 					if(ref($rule_value) eq 'ARRAY') {
+						my $ok = 1;
 						if(($rules->{'type'} eq 'integer') || ($rules->{'type'} eq 'number') || ($rules->{'type'} eq 'float')) {
 							unless(List::Util::any { $_ == $value } @{$rule_value}) {
-								if($rules->{'error_message'}) {
-									_error($logger, $rules->{'error_message'});
-								} else {
-									_error($logger, "validate_strict: Parameter '$key' ($value) must be one of ", join(', ', @{$rule_value}));
-								}
+								$ok = 0;
 							}
 						} else {
 							unless(List::Util::any { $_ eq $value } @{$rule_value}) {
-								if($rules->{'error_message'}) {
-									_error($logger, $rules->{'error_message'});
-								} else {
-									_error($logger, "validate_strict: Parameter '$key' ($value) must be one of ", join(', ', @{$rule_value}));
-								}
+								$ok = 0;
+							}
+						}
+
+						if(!$ok) {
+							if($rules->{'error_message'}) {
+								_error($logger, $rules->{'error_message'});
+							} else {
+								_error($logger, "validate_strict: Parameter '$key' ($value) must be one of ", join(', ', @{$rule_value}));
 							}
 						}
 					} else {
@@ -722,7 +962,7 @@ sub validate_strict
 									if($rules->{'error_message'}) {
 										_error($logger, $rules->{'error_message'});
 									} else {
-										_error($logger, "$key can only contain numbers (found $member)");
+										_error($logger, "$key can only contain integers (found $member)");
 									}
 								}
 							} elsif(($rule_value eq 'number') || ($rule_value eq 'float')) {
@@ -746,6 +986,8 @@ sub validate_strict
 					# Handled earlier
 				} elsif($rule_name eq 'error_message') {
 					# Handled inline
+				} elsif($rule_name eq 'transform') {
+					# Handled before the loop
 				} elsif($rule_name eq 'schema') {
 					# Nested schema Run the given schema against each element of the array
 					if($rules->{'type'} eq 'arrayref') {
@@ -759,7 +1001,9 @@ sub validate_strict
 					} elsif($rules->{'type'} eq 'hashref') {
 						if(ref($value) eq 'HASH') {
 							if(scalar keys(%{$value})) {
-								validate_strict({ input => $value, schema => $rule_value });
+								if(my $new_args = validate_strict({ input => $value, schema => $rule_value })) {
+									$value = $new_args;
+								}
 							}
 						} else {
 							_error($logger, "validate_strict: nested schema: Parameter '$value' must be an hashref");
